@@ -21,16 +21,18 @@ from celery.app.task import Task
 from celery.local import PromiseProxy
 from celery.signals import task_prerun, task_postrun, task_revoked
 from celery.utils.log import get_task_logger, get_logger
-from firexkit.revoke import revoke_recursively
 from firexkit.bag_of_goodies import BagOfGoodies
 from firexkit.argument_conversion import ConverterRegister
 from firexkit.result import get_tasks_names_from_results, wait_for_any_results, \
     RETURN_KEYS_KEY, wait_on_async_results_and_maybe_raise, get_result_logging_name, ChainInterruptedException, \
-    ChainRevokedException, extract_and_filter, last_causing_chain_interrupted_exception
+    ChainRevokedException, extract_and_filter, last_causing_chain_interrupted_exception, \
+    wait_for_running_tasks_from_results
 from firexkit.resources import get_firex_css_filepath, get_firex_logo_filepath
 from firexkit.firexkit_common import JINJA_ENV
 import time
 
+REPLACEMENT_TASK_NAME_POSTFIX = '_orig'
+
 logger = get_task_logger(__name__)
 
 
@@ -57,6 +59,10 @@ class DyanmicReturnsNotADict(Exception):
     pass
 
 
+class IllegalTaskNameException(Exception):
+    pass
+
+
 def create_collapse_ops(flex_collapse_ops_spec):
     from typing import Pattern
 
@@ -167,6 +173,11 @@ class FireXTask(Task):
     RETURN_KEYS_KEY = RETURN_KEYS_KEY
 
     def __init__(self):
+
+        check_name_for_override_posfix = getattr(self, 'check_name_for_override_posfix', True)
+        if check_name_for_override_posfix and self.name and self.name.endswith(REPLACEMENT_TASK_NAME_POSTFIX):
+            raise IllegalTaskNameException(f'Task names should never end with {REPLACEMENT_TASK_NAME_POSTFIX!r}')
+
         self.undecorated = undecorate(self)
         self.sig = inspect.signature(self.run)
         self._task_return_keys = self.get_task_return_keys()
@@ -189,6 +200,31 @@ class FireXTask(Task):
 
         self._from_plugin = False
 
+    @property
+    def root_orig(self):
+        """Return the very original `Task` that this `Task` had overridden.
+        If this task has been overridden multiple times, this will return the very first/original task.
+        Return `self` if the task was not overridden"""
+        if hasattr(self, "orig"):
+            return self.orig.root_orig
+        return self
+
+    def apply_async(self, *args, **kwargs):
+        original_name = self.name
+        if self.from_plugin and not original_name.endswith(REPLACEMENT_TASK_NAME_POSTFIX):
+            # If the task is overridden, and is not an intermediate override, then
+            # let's use the original name for serialization, in case that
+            # override name isn't available in the execution context.
+            # This can obviously be dangerous (but a risk we're deliberately taking)
+            # since we bind the args/kwargs/runtime options with the ovverriden service
+            # but might end up executing in a context that doesn't have it.
+            self.name = self.root_orig.name_without_orig
+        try:
+            res = super(FireXTask, self).apply_async(*args, **kwargs)
+        finally:
+            # Restore the original name
+            self.name = original_name
+        return res
 
     def signature(self, *args, **kwargs):
         # We need to lookup the task, in case it was over-ridden by a plugin
@@ -200,7 +236,6 @@ class FireXTask(Task):
             # These tests should really run in forked processes (or use Celery PyTest fixtures)
             # Otherwise, seems that everything is global
             new_self = self.app.tasks[self.name]
-
         # Get the signature from the new_self
         return super(FireXTask, new_self).signature(*args, **kwargs)
 
@@ -308,7 +343,7 @@ class FireXTask(Task):
 
     @staticmethod
     def strip_orig_from_name(task_name):
-        return re.sub("(_orig)*$", "", task_name)
+        return re.sub(f"({REPLACEMENT_TASK_NAME_POSTFIX})*$", "", task_name)
 
     @staticmethod
     def get_short_name(task_name):
@@ -329,7 +364,7 @@ class FireXTask(Task):
 
     @property
     def called_as_orig(self):
-        return True if self.name.endswith('_orig') else False
+        return True if self.name.endswith(REPLACEMENT_TASK_NAME_POSTFIX) else False
 
     @abstractmethod
     def pre_task_run(self, extra_events: Optional[dict] = None):
@@ -460,7 +495,7 @@ class FireXTask(Task):
         finally:
             try:
                 if self._lagging_children_strategy is not PendingChildStrategy.Continue:
-                    self.revoke_pending_children()
+                    self.revoke_nonready_children()
             finally:
                 self.remove_task_logfile_handler()
 
@@ -615,6 +650,10 @@ class FireXTask(Task):
         return [child for child, result in self.context.enqueued_children.items() if
                 result.get(self._STATE_KEY) == self._PENDING]
 
+    @property
+    def nonready_enqueued_children(self):
+        return [child for child in self.context.enqueued_children if not child.ready()]
+
     def _add_enqueued_child(self, child_result):
         if child_result not in self.context.enqueued_children:
             self.context.enqueued_children[child_result] = {}
@@ -636,7 +675,7 @@ class FireXTask(Task):
         child_results = self.pending_enqueued_children if pending_only else self.enqueued_children
         self.wait_for_specific_children(child_results=child_results, **kwargs)
 
-    def wait_for_specific_children(self, child_results, **kwargs: dict):
+    def wait_for_specific_children(self, child_results, **kwargs):
         """Wait for the explicitly provided child_results to run and complete"""
         if isinstance(child_results, AsyncResult):
             child_results = [child_results]
@@ -844,16 +883,31 @@ class FireXTask(Task):
             self.wait_for_specific_children(promises, raise_exception_on_failure=raise_exception_on_failure)
         return promises
 
-    def revoke_pending_children(self, **kwargs):
-        pending_children = self.pending_enqueued_children
-        if pending_children:
-            logger.info('Pending children of current task exist.')
-            [self.revoke_child(child_result, **kwargs) for child_result in pending_children]
+    def revoke_nonready_children(self):
+        nonready_children = self.nonready_enqueued_children
+        if nonready_children:
+            logger.info('Nonready children of current task exist.')
+            revoked = [self.revoke_child(child_result) for child_result in nonready_children]
+            wait_for_running_tasks_from_results([result for result_list in revoked for result in result_list])
 
-    def revoke_child(self, result: AsyncResult, **kwargs):
-        logger.debug('Revoking child %s' % get_result_logging_name(result))
-        revoke_recursively(result, **kwargs)
+    def revoke_child(self, result: AsyncResult, terminate=True, wait=False, timeout=None):
+        name = get_result_logging_name(result)
+        logger.debug('Revoking child %s' % name)
+        result.revoke(terminate=terminate, wait=wait, timeout=timeout)
+        revoked_results = [result]
         self._update_child_state(result, self._UNBLOCKED)
+        logger.info(f'Revoked {name}')
+
+        while result.parent:
+            # Walk up the chain, since nobody is waiting on those tasks explicitly.
+            result = result.parent
+            if not result.ready():
+                name = get_result_logging_name(result)
+                logger.debug(f'Revoking parent {name}')
+                result.revoke(terminate=terminate, wait=wait, timeout=timeout)
+                revoked_results.append(result)
+                logger.info(f'Revoked {name}')
+        return revoked_results
 
     @property
     def root_logger(self):
@@ -1224,4 +1278,4 @@ def statsd_task_postrun(sender, task, task_id, args, kwargs, **donotcare):
 
 @task_revoked.connect
 def statsd_task_revoked(sender, request, terminated, signum, expired, **kwargs):
-    send_task_completed_event(sender, request.id, sender.backend)
\ No newline at end of file
+    send_task_completed_event(sender, request.id, sender.backend)
